;;; mix-mode.el --- major mode for editing MIX code

;; Copyright (C) 2000 Free Software Foundation, Inc.

;; Authors: 2000 Sergey Poznyakoff
;; Version:  1.0
;; Keywords: mix TAOCP languages
;; $Id: mix-mode.el,v 1.2 2005/04/09 18:16:50 gray Exp $

;; This file is part of MIX simulator.

;; MIX simulator is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; MIX simulator is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with MIX simulator; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

(defvar mix-tab-stop-list '(8 13 32 40 48 56 64 72 80 88 96 104 112 120))

(defvar mix-mode-syntax-table nil
  "Syntax table used in mix-mode buffers.")
(if mix-mode-syntax-table
    ()
  (setq mix-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?* "<")
  (modify-syntax-entry ?\n ">"))
  

(defvar mix-mode-map ()
  "Keymap used in mix-mode buffers.")
(if mix-mode-map
    ()
  (setq mix-mode-map (make-sparse-keymap))
  (define-key mix-mode-map "\t" 'mix-indent-command)
  (define-key mix-mode-map "%" 'mix-match-paren-or-self-insert))
  
(defun mix-indent-command (&optional whole-exp)
  "Insert spaces or tabs to next defined tab-stop column in MIX mode.
The variable `mix-tab-stop-list' is a list of columns at which there
are tab stops. The variable `tab-stop-list' is ignored."
  (interactive)
  (and abbrev-mode (= (char-syntax (preceding-char)) ?w)
       (expand-abbrev))
  (let ((tabs mix-tab-stop-list))
    (while (and tabs (>= (current-column) (car tabs)))
      (setq tabs (cdr tabs)))
    (if tabs
	(let ((opoint (point)))
	  (skip-chars-backward " \t")
	  (delete-region (point) opoint)
	  (indent-to (car tabs)))
      (insert ?\ ))))

(defvar mix-mode-abbrev-table nil
  "Abbrev table in use in mix-mode buffers.")
(define-abbrev-table 'mix-mode-abbrev-table ())

;;; Label lookup routines

;; Find the definition of local label argH. Forward search.
(defun mix-find-local-forward (arg)
  (push-mark)
  (cond
   ((search-forward-regexp (concat "^" arg "H") nil t)
    (beginning-of-line))
   (t
    (pop-mark)
    (message (concat "Local definition of " arg "H not found")))))
  
;; Find the definition of local label argH. Backward search.
(defun mix-find-local-backward (arg)
  (push-mark)
  (cond
   ((search-backward-regexp (concat "^" arg "H") nil t)
    (beginning-of-line))
   (t
    (pop-mark)
    (message (concat "Local definition of " arg "H not found")))))

;; Find the reference to the local label argB. 
(defun mix-find-local-ref-forward (arg)
  (push-mark)
  (cond
   ((search-forward-regexp (concat "\\s " arg "B\\s ") nil t)
    (forward-char 1))
   (t
    (pop-mark)
    (message (concat "Reference to " arg "B not found")))))

;; Find the reference to the local label argF
(defun mix-find-local-ref-backward (arg)
  (push-mark)
  (cond
   ((search-backward-regexp (concat "\\s " arg "F\\s ") nil t)
    (forward-char 1))
   (t
    (pop-mark)
    (message (concat "Reference to " arg "F not found")))))

;; Find the definition of the label STRING, search forward
(defun mix-find-label-forward (string &optional msg)
  (push-mark)
  (forward-line 1)
  (cond
   ((search-forward-regexp (concat "^" string "\\s ") nil t)
    (forward-word -1))
   (msg
    (exchange-point-and-mark)
    (pop-mark)
    (message (concat "Label " string " not found")))
   (t
    (exchange-point-and-mark)
    (pop-mark)
    nil)))

;; Find the definition of the label STRING, search backward
(defun mix-find-label-backward (string &optional msg)
  (push-mark)
  (forward-line -1)
  (cond
   ((search-backward-regexp (concat "^" string "\\s ") nil t)
    t)
   (msg
    (exchange-point-and-mark)
    (pop-mark)
    (message (concat "Label " string " not found")))
   (t
    (exchange-point-and-mark)
    (pop-mark)
    nil)))

(defun mix-find-equ (string)
  (push-mark)
  (cond
   ((search-backward-regexp (concat "^" string "\\s EQU") nil t)
    t)
   (t
    (message (concat "Literal " string " not defined")))))

;; Ask user for the search direction 
(defun mix-read-search-direction ()
  "Ask user for the direction of the search.
Allowed replies are `f' for forward, `b' for backward and C-g for cancel."
  (message "search direction (f, b): ")
  (let (r)
    (while (not r)
      (let ((c (read-char)))
	(setq r (cond
		 ((eq c ?b)
		  "backward")
		 ((eq c ?f)
		  "forward")
		 (t
		  (progn
		    (message "Please, answer f or b:")
		    nil))))))
    r))

(defun mix-match-paren-hook (&optional arg)
 (interactive "p")
 (cond
  ((looking-at "\\([1-9]\\)B\\s ")
   (mix-find-local-backward (buffer-substring (match-beginning 1)
					 (1+ (match-beginning 1)))))
  ((looking-at "\\([1-9]\\)F\\s ")
   (mix-find-local-forward (buffer-substring (match-beginning 1)
	  			         (1+ (match-beginning 1)))))
  ((looking-at "^\\([1-9]\\)H\\s ")
   (let ((dir (mix-read-search-direction)))
     (funcall (intern (concat "mix-find-local-ref-" dir))
	      (buffer-substring (match-beginning 1)
				(1+ (match-beginning 1))))))
  ((looking-at "\\s *\\([A-Z][A-Z0-9]*\\)")
   (or
    (mix-find-label-forward (buffer-substring (match-beginning 1)
					      (match-end 1)))
    (mix-find-label-backward (buffer-substring (match-beginning 1)
					       (match-end 1)) t)))
  ((looking-at "=\\([A-Z][A-Z0-9]*\\)=")
   (mix-find-equ (buffer-substring (match-beginning 1)
				   (match-end 1))))
  (arg
   (self-insert-command arg))))

(defun mix-match-paren-or-self-insert (arg)
  (interactive "p")
  (mix-match-paren-hook arg))

(defun mix-match-paren ()
  (mix-match-paren-hook))

;;;###autoload
(defun mix-mode ()
  "Major mode for editing MIX code.

Key bindings:
\\{mix-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (set-syntax-table mix-mode-syntax-table)
  (setq major-mode 'mix-mode
	mode-name "MIX"
	local-abbrev-table mix-mode-abbrev-table)
  (use-local-map mix-mode-map))
  
(provide 'mix-mode)
;;; mix-mode ends

